# Copyright (C) 2015 Stefan C. Mueller

import unittest
from pydron_dataflow.dataflow import graph
from pydron_dataflow.dataflow.graph import START_TICK, G, T, C
from pydron_dataflow.dataflow import utils
import pickle

class TestTick(unittest.TestCase):
    
    def test_inc(self):
        inc = graph.START_TICK + 1
        self.assertLess(graph.START_TICK, inc)
        
    def test_inc_2(self):
        inc1 = graph.START_TICK + 1
        inc2 = graph.START_TICK + 2
        self.assertLess(inc1, inc2)
        
    def test_inc_inc(self):
        inc1 = graph.START_TICK + 1
        inc2 = inc1 + 1
        self.assertLess(inc1, inc2)
        
    def test_inc_loop(self):
        t = graph.Tick((0,4), (False, True))
        t += 10
        expected = graph.Tick((0,14), (False, True))
        self.assertEqual(t, expected)
        
    def test_shift_less_than_base(self):
        a = graph.START_TICK
        b = a + 1
        self.assertLess(a, graph.START_TICK << b)
        
    def test_shift_more_than_next(self):
        a = graph.START_TICK
        b = a + 1
        c = a + 2
        self.assertLess(graph.START_TICK << b, c)
        
    def test_shift(self):
        a = graph.Tick((0,1), (False, False))
        b = graph.Tick((0,2), (False, False))
        actual = a << b
        expected = graph.Tick((0,2,1), (False, False, False))
        self.assertEqual(actual, expected)
        
    def test_shift_keeps_order(self):
        a = graph.START_TICK
        b = a + 1
        self.assertLess(graph.START_TICK << b, (graph.START_TICK << b) + 1)
        
    def test_shift_loop(self):
        a = graph.Tick((0,4), (False, True))
        b = graph.Tick((0,10,5), (False, True, False))
        
        actual = a << b
        expected = graph.Tick((0,10,5, 4), (False, True, False, True))
        self.assertEqual(actual, expected)
        
    def test_mark_loop_iteration(self):
        t = graph.Tick((0,10,5), (False, False, False))
        actual = t.mark_loop_iteration()
        expected = graph.Tick((0,10,5), (False, False, True))
        self.assertEqual(actual, expected)
        
    def test_non_loop_elements(self):
        t = graph.Tick((0,10,5,3), (False, True, False, True))
        self.assertEqual((0,5), t.nonloop_elements)
        
    def test_loop_elements(self):
        t = graph.Tick((0,10,5,3), (False, True, False, True))
        self.assertEqual((10, 3), t.loop_elements)
        
    def test_repr_start(self):
        self.assertEqual("START_TICK", repr(graph.START_TICK))
        
    def test_repr_final(self):
        self.assertEqual("FINAL_TICK", repr(graph.FINAL_TICK))
        
    def test_repr_noshift(self):
        self.assertEqual("4", repr(graph.START_TICK + 4))
        
    def test_repr_shift(self):
        self.assertEqual("(10, 4)", repr(graph.START_TICK + 4 << graph.START_TICK + 10))
        
    def test_repr_noshift_loop(self):
        t = graph.Tick((0,4), (False, True))
        self.assertEqual("*4", repr(t))
        
    def test_repr_shift_loop(self):
        t = graph.Tick((0,10,4), (False, True, False))
        self.assertEqual("(*10, 4)", repr(t))
        
    
    def test_parse_start(self):
        self.assertEqual(graph.START_TICK, graph.Tick.parse_tick("START_TICK"))

    def test_parse_final(self):
        self.assertEqual(graph.FINAL_TICK, graph.Tick.parse_tick("FINAL_TICK"))
        
    def test_parse_int(self):
        self.assertEqual(graph.START_TICK + 42, graph.Tick.parse_tick(42))
        
    def test_parse_tuple(self):
        self.assertEqual(graph.START_TICK + 2 << graph.START_TICK + 42, graph.Tick.parse_tick((42, 2)))
        
    def test_parse_str(self):
        expected = graph.Tick((0,1,2,3), (False, False, False, False))
        actual = graph.Tick.parse_tick("1,2,3")
        self.assertEqual(actual, expected)
        
    def test_parse_str_loop(self):
        expected = graph.Tick((0,1,2,3), (False, False, True, False))
        actual = graph.Tick.parse_tick("1,*2,3")
        self.assertEqual(actual, expected)
        
    def assertEqual(self, first, second, msg=None):
        unittest.TestCase.assertEqual(self, first, second, msg=msg)
        if isinstance(first, graph.Tick) and isinstance(second, graph.Tick):
            unittest.TestCase.assertEqual(self, first._loopmask, second._loopmask, msg=msg)
        
    def assertEquals(self, first, second, msg=None):
        self.assertEqual(first, second, msg)
        
class TestGraph(unittest.TestCase):
    
    def setUp(self):
        self.target = graph.Graph()
        self.observer = MockObserver()
    
    def test_pickle(self):
        self.target.add_task(START_TICK + 100, None, {'nicename':'test1'})
        self.target.add_task(START_TICK + 101, None, {'nicename':'test2'})
        self.target.connect(graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))
        
        s = pickle.dumps(self.target, protocol=pickle.HIGHEST_PROTOCOL)
        copy = pickle.loads(s)
        self.assertEqual(self.target, copy)
    
    def test_add_task(self):
        self.target.add_task(START_TICK + 100, None, {'nicename':'test'})
        self.assertEqual([START_TICK + 100], list(self.target.get_all_ticks()))
        
    def test_remove_task(self):
        self.target.add_task(START_TICK + 100, None, {'nicename':'test'})
        self.target.remove_task(START_TICK + 100)
        self.assertEqual([], list(self.target.get_all_ticks()))
        
    def test_get_task_properies(self):
        self.target.add_task(START_TICK + 100, None, {'nicename':'test'})
        self.assertEqual({'nicename':'test'}, self.target.get_task_properties(START_TICK + 100))
        
    def test_set_task_property(self):
        self.target.add_task(START_TICK + 100, None, {'nicename':'test'})
        self.target.set_task_property(START_TICK + 100, 'syncpoint', True)
        self.assertEqual({'nicename':'test', 'syncpoint': True}, self.target.get_task_properties(START_TICK + 100))
        
    def test_get_in_connections(self):
        self.target.add_task(START_TICK + 100, None, {'nicename':'test1'})
        self.target.add_task(START_TICK + 101, None, {'nicename':'test2'})
        
        self.target.connect(graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))
        
        actual = list(self.target.get_in_connections(START_TICK + 101))
        
        expected = [(graph.Endpoint(START_TICK + 100, 'out'), graph.Endpoint(START_TICK + 101, 'in'))]
        self.assertEqual(actual, expected)
        
    def test_get_out_connections(self):
        self.target.add_task(START_TICK + 100, None, {'nicename':'test1'})
        self.target.add_task(START_TICK + 101, None, {'nicename':'test2'})
        
        self.target.connect(graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))
        
        actual = list(self.target.get_out_connections(START_TICK + 100))
        
        expected = [(graph.Endpoint(START_TICK + 100, 'out'), graph.Endpoint(START_TICK + 101, 'in'))]
        self.assertEqual(actual, expected)
        
    def test_disconnect(self):
        self.target.add_task(START_TICK + 100, None, {'nicename':'test1'})
        self.target.add_task(START_TICK + 101, None, {'nicename':'test2'})
        
        self.target.connect(graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))
        
        self.target.disconnect(graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))
        
        actual = list(self.target.get_in_connections(START_TICK + 101))
        expected = []
        self.assertEqual(actual, expected)
        
    def test_same_tick(self):
        self.target.add_task(START_TICK + 100, None, {'nicename':'test1'})
        self.assertRaises(ValueError, self.target.add_task, START_TICK + 100, {'nicename':'test2'})
        
    def test_causality(self):
        self.target.add_task(START_TICK + 100, None, {'nicename':'test1'})
        self.target.add_task(START_TICK + 101, None, {'nicename':'test2'})
        
        self.assertRaises(ValueError, self.target.connect, 
                          graph.Endpoint(START_TICK + 101, 'out'), 
                          graph.Endpoint(START_TICK + 100, 'in'))
        
    def test_double_input(self):
        
        self.target.add_task(START_TICK + 100, None, {'nicename':'test1'})
        self.target.add_task(START_TICK + 101, None, {'nicename':'test2'})
        self.target.add_task(START_TICK + 102, None, {'nicename':'test3'})
        self.target.connect(graph.Endpoint(START_TICK + 100, 'out'), 
                                          graph.Endpoint(START_TICK + 101, 'in'))
        
        self.assertRaises(ValueError, self.target.connect, 
                          graph.Endpoint(START_TICK + 101, 'out'), graph.Endpoint(START_TICK + 101, 'in'))
        
    def test_get_task(self):
        task = object()
        self.target.add_task(START_TICK + 100, task)
        self.assertEqual(task, self.target.get_task(START_TICK + 100))
        
    def test_mocking(self):
        self.target.add_task(START_TICK + 100, "task1", {'nicename':'test1'})
        self.target.add_task(START_TICK + 101, "task2", {'nicename':'test2'})
        self.target.connect(graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))
        
        expected =  G(
            T(100, 'task1', {'nicename': 'test1'}),
            C(100, 'out', 101, 'in'),
            T(101, 'task2', {'nicename': 'test2'}),
        )

        utils.assert_graph_equal(expected, self.target)
        
    def test_observer_add_task(self):
        self.target.subscribe(self.observer)
        self.target.add_task(START_TICK + 100, "task1", {'nicename':'test1'})
        self.assertEqual([("task_added", START_TICK + 100, "task1", {'nicename':'test1'})], self.observer.calls)
        
    def test_observer_remove_task(self):
        self.target.add_task(START_TICK + 100, "task1")
        self.target.subscribe(self.observer)
        self.target.remove_task(START_TICK + 100)
        self.assertEqual([("task_removed", START_TICK + 100)], self.observer.calls)
        
    def test_observer_connect(self):
        self.target.add_task(START_TICK + 100, "task1")
        self.target.add_task(START_TICK + 101, "task2")
        self.target.subscribe(self.observer)
        self.target.connect(graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))
        self.assertEqual([("connected", graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))], self.observer.calls)
        
    def test_observer_disconnect(self):
        self.target.add_task(START_TICK + 100, "task1")
        self.target.add_task(START_TICK + 101, "task2")
        
        self.target.connect(graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))
        self.target.subscribe(self.observer)
        self.target.disconnect(graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))
        self.assertEqual([("disconnected", graph.Endpoint(START_TICK + 100, 'out'), 
                            graph.Endpoint(START_TICK + 101, 'in'))], self.observer.calls)
        
    def test_observer_set_task_property(self):
        self.target.add_task(START_TICK + 100, "task1")
        self.target.subscribe(self.observer)
        self.target.set_task_property(START_TICK + 100, "foo", "bar")
        self.assertEqual([("task_property_changed", START_TICK + 100, "foo", "bar")], self.observer.calls)
        
class MockObserver(object):
    
    def __init__(self):
        self.calls = []
        
    def __getattr__(self, name):
        def mock(*args):
            self.calls.append((name,) + args)
        return mock



# Copyright (C) 2015 Stefan C. Mueller

import unittest

from pydron_dataflow.interpreter import graphdecorator
from pydron_dataflow.dataflow import graph


TICK1 = graph.START_TICK + 1
TICK2 = graph.START_TICK + 2
TICK3 = graph.START_TICK + 3
TICK4 = graph.START_TICK + 4
TICK5 = graph.START_TICK + 5
FINAL = graph.FINAL_TICK

class TestDataDecorator(unittest.TestCase):
    
    def setUp(self):
        self.target = graphdecorator.DataGraphDecorator(graph.Graph())

    def test_get_data(self):
        self.target.add_task(graph.START_TICK + 1, "task1")
        self.target.set_output_data(graph.START_TICK + 1, {"out1": "data1", "out2":"data2"})
        actual = self.target.get_data(graph.Endpoint(graph.START_TICK + 1, "out1"))
        self.assertEqual("data1", actual)
        
    def test_get_nonexistent_data(self):        
        self.target.add_task(graph.START_TICK + 1, "task1")
        self.target.set_output_data(graph.START_TICK + 1, {"out1": "data1", "out2":"data2"})
        self.assertRaises(KeyError, self.target.get_data, graph.Endpoint(graph.START_TICK + 1, "out3"))
        
    def no_data(self):
        self.target.add_task(TICK1, "task1")
        self.assertRaises(KeyError, self.target.get_data, graph.Endpoint(TICK1, "out"))
                
    def test_graph_input(self):
        self.target.add_task(TICK1, MockTask("in"))
        self.target.connect(graph.Endpoint(graph.START_TICK, "graphin"), graph.Endpoint(TICK1, "in"))
        self.assertRaises(KeyError, self.target.get_data, graph.Endpoint(graph.START_TICK, "graphin"))
        
class TestReadyDecorator(unittest.TestCase):
    
    
    def setUp(self):
        self.target = graphdecorator.ReadyDecorator(
                            graphdecorator.DataGraphDecorator(graph.Graph()))
        
    def test_add_task_flush(self):
        self.target.add_task(TICK1, "task")
        
        actual = self.target.collect_ready_tasks()
        expected = {TICK1, FINAL}
        
        self.assertEqual(actual, expected)
        
    def test_remove_task_after_flush(self):
        self.target.add_task(TICK1, "task")
        
        self.target.collect_ready_tasks()
        self.target.remove_task(TICK1)
        
        actual = self.target.collect_ready_tasks()
        expected = set()
        
        self.assertEqual(actual, expected)
        
    def test_add_conn_before_flush(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, "task2")
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))

        actual = self.target.collect_ready_tasks()
        expected = {FINAL, TICK1}
        
        self.assertEqual(actual, expected)
        
    def test_disconnect_before_flush(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, "task2")
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        self.target.disconnect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        
        actual = self.target.collect_ready_tasks()
        expected = {TICK1, TICK2, FINAL}
        
        self.assertEqual(actual, expected)
        
    def test_add_conn_after_flush(self):
        self.target.add_task(graph.START_TICK + 1, "task1")
        self.target.add_task(graph.START_TICK + 2, "task2")
        
        self.target.collect_ready_tasks()
        
        self.target.connect(graph.Endpoint(graph.START_TICK + 1, "out"), 
                          graph.Endpoint(graph.START_TICK + 2, "in"))
        self.target.set_output_data(graph.START_TICK + 1, {"out": "data"})
        
        self.assertEqual(set(), self.target.collect_ready_tasks())
        
    
    def test_set_output_data(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, "task2")
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))

        self.target.set_output_data(TICK1, {"out": "data"})
        
        actual = self.target.collect_ready_tasks()
        expected = {TICK1, TICK2, FINAL}
        
        self.assertEqual(actual, expected)
        
    def test_set_output_data_final(self):
        self.target.add_task(TICK1, "task1")
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(FINAL, "in"))
        
        self.assertEqual({TICK1}, self.target.collect_ready_tasks())
        
        self.target.set_output_data(TICK1, {"out": "data"})
        
        self.assertEqual({FINAL}, self.target.collect_ready_tasks())
        
        
    def test_set_data_before_connection(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, "task2")
        self.target.set_output_data(TICK1, {"out": "data"})
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        actual = self.target.collect_ready_tasks()
        expected = {TICK1, TICK2, FINAL}
        self.assertEqual(actual, expected)
        
    def test_overwrite_data(self):
        self.target.add_task(graph.START_TICK + 1, "task1")
        self.target.set_output_data(graph.START_TICK + 1, {"out1": "data1", "out2":"data2"})
        self.assertRaises(ValueError, self.target.set_output_data, graph.START_TICK + 1, {"out2":"data2"})
        
    def test_syncpoint_only_before(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, "task2")
        self.target.add_task(TICK3, "task3", {"syncpoint":True})
        self.target.add_task(TICK4, "task4")
        self.target.add_task(TICK5, "task4")
        self.assertEqual({TICK1, TICK2}, self.target.collect_ready_tasks())
        
    def test_syncpoint_alone(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, "task2")
        self.target.add_task(TICK3, "task3", {"syncpoint":True})
        self.target.add_task(TICK4, "task4")
        self.target.add_task(TICK5, "task4")
        
        self.target.collect_ready_tasks()
        self.target.set_output_data(TICK1, {})
        self.target.set_output_data(TICK2, {})
        
        self.assertEqual({TICK3}, self.target.collect_ready_tasks())
        
    def test_syncpoint_after(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, "task2")
        self.target.add_task(TICK3, "task3", {"syncpoint":True})
        self.target.add_task(TICK4, "task4")
        self.target.add_task(TICK5, "task4")
        
        self.target.collect_ready_tasks()
        self.target.set_output_data(TICK1, {})
        self.target.set_output_data(TICK2, {})
        self.target.collect_ready_tasks()
        self.target.set_output_data(TICK3, {})
        
        self.assertEqual({TICK4, TICK5, FINAL}, self.target.collect_ready_tasks())
        
    def test_syncpoint_remove(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, "task2")
        self.target.add_task(TICK3, "task3", {"syncpoint":True})
        self.target.add_task(TICK4, "task4")
        self.target.add_task(TICK5, "task4")
        self.target.set_task_property(TICK3, "syncpoint", False)
        self.assertEqual({TICK1, TICK2, TICK3, TICK4, TICK5, FINAL}, self.target.collect_ready_tasks())
        
    def test_unrefined(self):
        self.target.add_task(TICK1, MockTask("in"))
        self.assertEqual({FINAL}, self.target.collect_ready_tasks())
        
    def test_refined_initially(self):
        self.target.add_task(TICK1, MockTask("in"), {"refined": True})
        self.assertEqual({TICK1, FINAL}, self.target.collect_ready_tasks())
        
    def test_refined_later(self):
        self.target.add_task(TICK1, MockTask("in"))
        self.assertEqual({FINAL}, self.target.collect_ready_tasks())
        self.target.set_task_property(TICK1, "refined", True)
        self.assertEqual({TICK1}, self.target.collect_ready_tasks())
        
    def test_graph_input(self):
        self.target.add_task(TICK1, MockTask("in"))
        self.target.connect(graph.Endpoint(graph.START_TICK, "graphin"), graph.Endpoint(TICK1, "in"))
        self.assertEqual({FINAL}, self.target.collect_ready_tasks())
        
    def test_was_collected_inqueue(self):
        self.target.add_task(TICK1, "task")
        self.assertFalse(self.target.was_ready_collected(TICK1))
        
    def test_was_collected_true(self):
        self.target.add_task(TICK1, "task")
        self.target.collect_ready_tasks()
        self.assertTrue(self.target.was_ready_collected(TICK1))

    def test_was_collected_not_ready(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, "task2")
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        self.assertFalse(self.target.was_ready_collected(TICK2))
        
class TestRefineDecorator(unittest.TestCase):
    
    
    def setUp(self):
        self.target = graphdecorator.RefineDecorator(
                            graphdecorator.DataGraphDecorator(graph.Graph()))
        
    def test_add_task_disconnected(self):
        self.target.add_task(TICK1, MockTask("in"))
        self.assertEqual(set(), self.target.collect_refine_tasks())
        
    def test_connect_no_data(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, MockTask("in"))
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        self.assertEqual(set(), self.target.collect_refine_tasks())
        
    def test_set_data_after_connect(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, MockTask("in"))
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        self.target.set_output_data(TICK1, {"out": "data"})
        self.assertEqual({TICK2}, self.target.collect_refine_tasks())
        
    def test_set_data_before_connect(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, MockTask("in"))
        self.target.set_output_data(TICK1, {"out": "data"})
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        self.assertEqual({TICK2}, self.target.collect_refine_tasks())
        
    def test_disconnect(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, MockTask("in"))
        self.target.set_output_data(TICK1, {"out": "data"})
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        self.target.disconnect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        self.assertEqual(set(), self.target.collect_refine_tasks())
        
    def test_disconnect_two(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, MockTask("in1", "in2"))
        self.target.set_output_data(TICK1, {"out": "data"})
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in1"))
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in2"))
        self.target.disconnect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in1"))
        self.assertEqual({TICK2}, self.target.collect_refine_tasks())

    def test_was_collected_inqueue(self):
        self.target.add_task(TICK1, "task")
        self.assertFalse(self.target.was_refine_collected(TICK1))
        
    def test_was_collected_no_refiner_ports(self):
        self.target.add_task(TICK1, "task")
        self.target.collect_refine_tasks()
        self.assertFalse(self.target.was_refine_collected(TICK1))
        
    def test_was_collected_true(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, MockTask("in"))
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        self.target.set_output_data(TICK1, {"out": "data"})
        self.target.collect_refine_tasks()
        self.assertTrue(self.target.was_refine_collected(TICK2))

    def test_was_collected_not_ready(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, "task2")
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        self.assertFalse(self.target.was_refine_collected(TICK2))
        
    def test_will_be_refined(self):
        self.target.add_task(TICK1, "task1")
        self.target.add_task(TICK2, MockTask("in"))
        self.target.connect(graph.Endpoint(TICK1, "out"), graph.Endpoint(TICK2, "in"))
        self.assertTrue(self.target.will_be_refined(TICK2))
        
    def test_will_not_be_refined(self):
        self.target.add_task(TICK1, "task")
        self.assertFalse(self.target.will_be_refined(TICK1))
        
class MockTask(object):
    def __init__(self, *refiner_ports):
        self.refiner_ports = set(refiner_ports)


# Copyright (C) 2015 Stefan C. Mueller

import unittest
from pydron_dataflow.dataflow import refine, utils
from pydron_dataflow.dataflow.graph import G,T,C, START_TICK, FINAL_TICK

class TestReplaceTask(unittest.TestCase):
    
    def test_in_out(self):
        
        subgraph = G(
            C(START_TICK, "sin", 1, "x"),
            T(1, "subtask"),
            C(1, "y", FINAL_TICK, "sout")
        )
        
        g = G(
              C(START_TICK, "in", 1, "sin"),
              T(1, "task_to_replace"),
              C(1, "sout", FINAL_TICK, "out")
              )
        
        expected  = G(
              C(START_TICK, "in", (1,1), "x"),
              T((1,1), "subtask"),
              C((1,1), "y", FINAL_TICK, "out")
              )
        
        refine.replace_task(g, START_TICK + 1, subgraph)
        
        utils.assert_graph_equal(expected, g)
        
    def test_unassigned_out(self):
        
        subgraph = G(
            C(START_TICK, "sin", 1, "x"),
            T(1, "subtask"),
            C(1, "y", FINAL_TICK, "sout")
        )
        
        g = G(
              C(START_TICK, "in", 1, "sin"),
              C(START_TICK, "sout2", 1, "sout2"),
              T(1, "task_to_replace"),
              C(1, "sout", FINAL_TICK, "out"),
              C(1, "sout2", FINAL_TICK, "out2")
              )
        
        expected  = G(
              C(START_TICK, "in", (1,1), "x"),
              T((1,1), "subtask"),
              C((1,1), "y", FINAL_TICK, "out"),
              C(START_TICK, "sout2", FINAL_TICK, "out2")
              )
        
        refine.replace_task(g, START_TICK + 1, subgraph)
        
        utils.assert_graph_equal(expected, g)
        
        
    def test_passhtrough(self):
        
        subgraph = G(
            C(START_TICK, "sin", FINAL_TICK,"sout"),
        )
        
        g = G(
              C(START_TICK, "in", 1, "sin"),
              T(1, "task_to_replace"),
              C(1, "sout", FINAL_TICK, "out"),
              )
        
        expected  = G(
              C(START_TICK, "in", FINAL_TICK, "out"),
              )
        
        refine.replace_task(g, START_TICK + 1, subgraph)
        
        utils.assert_graph_equal(expected, g)
        


# Copyright (C) 2015 Stefan C. Mueller

"""
Methods that help with changeing the graph.
"""

from pydron_dataflow.dataflow import graph
    
def replace_task(g, tick, subgraph, subgraph_tick=None, additional_inputs={}):
    """
    Replaces the task in `g` at `tick` with `subgraph`.
    
    All tasks of the subgraph will be shifted by `subgraph_tick`. If `subgraph_tick` is `None`
    then the tasks are shifted by `tick`.
    
    All inputs of the subgraph are connected to source endpoints in `g`. Endpoints for input port
    `p` are searched as follows:
    
    * if `p` is a key in  `additional_inputs` then the value is expected to be the endpoint in `g`
      to use.
      
    * if the replaced task has an input port named `p`, then the endpoint which connects to that port
      is used.
      
    * if none of the above matches, fail with an exception.
    
    For each output port `p` of the replaced task, the destination endpoint in `g` will be connected 
    to another source endpoint
    
    * if the subgraph has an output `g` then it is connected to the source of that output.
      If the source of that output is an input to the subgraph, then the source is searched using
      the rules above for inputs of the subgraph.
      
    * use the rules for inputs to find the source. This will fail if there is none to be found.

    """
    
    if subgraph_tick is None:
        subgraph_tick = tick
    
    task_input_map = {dest.port: source for source, dest in g.get_in_connections(tick)}
    task_input_map.update(additional_inputs)
    
    subgraph_output_map = {dest.port: source for source, dest in subgraph.get_in_connections(graph.FINAL_TICK)}
    

    # Prepare the connections to hook up the subgraph's inputs.
    input_connections = []
    for source, dest in subgraph.get_out_connections(graph.START_TICK):
        task_input = task_input_map[source.port]
        if dest.tick == graph.FINAL_TICK:
            continue # direct connection are treated as output connecions.
        else:
            subgraph_dest = graph.Endpoint(dest.tick << subgraph_tick, dest.port)
            input_connections.append((task_input, subgraph_dest))
        
    # Prepare the connections to replace the ones of the removed task.
    output_connections = []
    for source, dest in g.get_out_connections(tick):
        if source.port in subgraph_output_map:
            # output is assigned in the subgraph
            source = subgraph_output_map[source.port]
            if source.tick == graph.START_TICK:
                # not connected to a task in the subgraph. This is a connection from
                # START_TICK to FINAL_TICK. We have to find the task input for this.
                source = task_input_map[source.port]
            else:
                source = graph.Endpoint(source.tick << subgraph_tick, source.port)
        elif source.port in task_input_map:
            # output is not assigned, but is an input to the replaced task.
            # Just pass it through.
            source = task_input_map[source.port]
        else:
            raise ValueError("No input on the replaced task for output %s" % repr(source.port))
        output_connections.append((source, dest))
    
    # Remove the task
    for source, dest in g.get_in_connections(tick):
        g.disconnect(source, dest)
    for source, dest in g.get_out_connections(tick):
        g.disconnect(source, dest)
    g.remove_task(tick)
    
    # Insert subgraph, make connections
    insert_subgraph(g, subgraph, subgraph_tick)
    for source, dest in input_connections:
        g.connect(source, dest)
    for source, dest in output_connections:
        g.connect(source, dest)
    

def insert_subgraph(g, subgraph, supertick):
    """
    Inserts all tasks and connections between them from `subgraph` into `g`.
    All ticks are shifted by `supertick`. The connections to START_TICK and FINAL_TICK
    are NOT copied.
    """
    
    for tick in subgraph.get_all_ticks():
        newtick = tick << supertick
        g.add_task(newtick, subgraph.get_task(tick), subgraph.get_task_properties(tick))
        
    
    for tick in list(subgraph.get_all_ticks()) + [graph.FINAL_TICK]:
        
        for source, dest in subgraph.get_in_connections(tick):
            if source.tick == graph.START_TICK or dest.tick == graph.FINAL_TICK:
                continue
            g.connect(graph.Endpoint(source.tick << supertick, source.port),
                      graph.Endpoint(dest.tick << supertick, dest.port))
        
        


# Copyright (C) 2015 Stefan C. Mueller

import ast

def contains_sideeffects(g):
    """
    Checks if the graph contains at least one task with
    a task property `syncpoint` which is `True`.
    
    It also checks subgraphs stored within tasks, given
    that the task objects have a `subgraphs()` method.
    """
    
    visited_graphs = set()
    
    def visit(g):
        
        # avoid infinite loops if subgraphs are cyclic.
        if id(g) in visited_graphs:
            return False
        visited_graphs.add(id(g))
        
        for tick in g.get_all_ticks():
            if g.get_task_properties(tick).get('syncpoint', False):
                return True
            
            task = g.get_task(tick)
            for subgraph in task.subgraphs():
                if visit(subgraph):
                    return True
        return False
    return visit(g)
    

def assert_graph_equal(expected, actual):
    
    def trim(lines):
        for line in lines:
            line = line.strip()
            if line:
                yield line
                
                
    if expected != actual:
        expected_clean = "\n".join(trim(repr(expected).splitlines()))
        actual_clean = "\n".join(trim(repr(actual).splitlines()))
    
        
        for i, (e, a) in enumerate(zip(expected_clean.splitlines(), actual_clean.splitlines())):
            if e != a:
                msg = "Difference on line %s: Expected %s got %s" % (i, repr(e), repr(a))
                break
        else:
            msg = "No difference in string representation."
        
        print(repr(expected))
        print(repr(actual))
        raise AssertionError("Graphs differ: %s" % msg)



# Copyright (C) 2015 Stefan C. Mueller

import sys
from sortedcontainers import SortedSet
from pydron_dataflow.dataflow import graph

class AbstractGraphDecorator(object):
    
    def __init__(self, g):
        if len(g.get_all_ticks()) != 0:
            raise ValueError("Graph must be empty.")
        self.g = g
        
    def add_task(self, tick, task, properties={}):
        self.g.add_task(tick, task, properties)
               
    def remove_task(self, tick):
        return self.g.remove_task(tick)
            
    def connect(self, source, dest):
        return self.g.connect(source, dest)
    
    def disconnect(self, source, dest):
        return self.g.disconnect(source, dest)

    def get_all_ticks(self):
        return self.g.get_all_ticks()
    
    def get_task(self, tick):
        return self.g.get_task(tick)
    
    def get_task_properties(self, tick):
        return self.g.get_task_properties(tick)
        
    def set_task_property(self, tick, key, value):
        return self.g.set_task_property(tick, key, value)
        
    def get_in_connections(self, tick):
        return self.g.get_in_connections(tick)
                
    def get_out_connections(self, tick):
        return self.g.get_out_connections(tick)
    
    def __getattr__(self, attr):
        return getattr(self.g, attr)
    
    def __repr__(self):
        return repr(self.g)
    
class DataGraphDecorator(AbstractGraphDecorator):
    """
    This decorator adds a task property called
    `out_data` to each task. The property contains
    a dict with a output-port to value mapping.
    
    Even ports that aren't currently connected
    may have data. Once the data for a particular
    port is set, it cannot be changed.
    """
    
    def __init__(self, g):
        AbstractGraphDecorator.__init__(self, g)
        self.set_task_property(graph.START_TICK, "out_data", {})

    def set_output_data(self, tick, outputs):
        """
        Set the output values of a task.
        Once an output is set, it cannot be changed.
        
        This also marks the task as 'executed'. It should
        therefore be invoked once the task has finished
        even if the task has no output ports.
        
        :param tick: Tick of the task that has completed.
        :param outputs: Dict with out-port to value-reference map. 
        """
        props = self.g.get_task_properties(tick)
        if (sys.version_info > (3, 0)):
            for port, data in outputs.items():
                if port in props["out_data"]:
                    raise ValueError("Value of out-port %s is already set." % port)
                props["out_data"][port] = data
        else:
            for port, data in outputs.iteritems():
                if port in props["out_data"]:
                    raise ValueError("Value of out-port %s is already set." % port)
                props["out_data"][port] = data
            
    def get_data(self, out_endpoint):
        """
        Returns the data reference for the specified output port.
        
        If the data reference was not set, an error is raised.
        If the data reference was set but was returned by :meth:`gc` in
        the mean time then an error is raised too.
        """
        props = self.get_task_properties(out_endpoint.tick)
        try:
            data = props["out_data"][out_endpoint.port]
            return data
        except KeyError:
            raise KeyError("Data is not available.")
        

            
    def add_task(self, tick, task, properties={}):
        properties["out_data"] = {}
        self.g.add_task(tick, task, properties=properties)
    
        
class AbstractReadyDecorator(AbstractGraphDecorator):
    """
    Wraps a graph and keeps track of tasks that are `ready`.
    
    A task is ready if and only if:
    
     * For each input connection to certain ports: The connected output port has data set with `set_outut_data()`.
       The input ports can be selected with the `port_filter` passed to `__init__`.
     
     * The task is not a sync-point and all sync-point tasks with a lower tick have been executed.
     
     * The task is a sync-point task and all tasks with a lower tick have been executed.
     
    A task is considered to be `executed` once `set_output_data()` has been called.
    """
    
    def __init__(self, g, prefix, port_filter, property_filter, syncpoint_run_last=True):
        """
        :param prefix: Prefix for the task properties used by the implementation
          to keep track of the state. If several decorators of this type
          are wrapped around the same graph, then different prefixes must be used.
        :param port_filter: Function to filter the relevant input ports. The function
          takes three arguments:
           * The graph (`self`)
           * The tick of the task.
           * The name of the port.
          If the function returns `False` Then the port is ignored.
        :param property_filter: Function that filters tasks which are collected.
          The function takes three
           * The graph (`self`)
           * The tick of the task.
           * The properties of the task (`graph.get_task_properties(tick)`).
          Only tasks for which this function returns `True` are returned. The function
          is reevaluated if the properties of a task change.
        :param syncpoint_run_last: If `True` then tasks with syncpoints only
          run once all tasks with a lower tick have completed.
        """
        AbstractGraphDecorator.__init__(self, g)
        self._prefix = prefix
        self._port_filter = port_filter
        self._property_filter = property_filter
        self._syncpoint_run_last = syncpoint_run_last
        
        #: ticks of all tasks that have data for all inputs.
        #: That is, if every output port connected to each input port
        #: had data set with `set_output_data`.
        self._queue = SortedSet()
        
        #: ticks of all tasks with
        #: * `properties["syncpoint"] == True`
        #: * `set_output_data` not yet called.
        self._pending_syncpoints = SortedSet()
        
        #: ticks of all tasks with
        #: * `set_output_data` not yet called.
        self._pending_ticks = SortedSet()
        
        self._collected_prop = self._prefix + "_collected"
        self._count_prop = self._prefix + "_count"
        self._ready_prop = self._prefix + "_ready"
        
        self.g.set_task_property(graph.FINAL_TICK, self._count_prop, 0)
        self.g.set_task_property(graph.FINAL_TICK, self._ready_prop, 0)
        self.g.set_task_property(graph.FINAL_TICK, self._collected_prop, False)
        self._consider(graph.FINAL_TICK)
        self._pending_ticks.add(graph.FINAL_TICK)
        
    def past_all_syncpoints(self):
        """
        Returns `True` if all sync-point tasks have been executed.
        """
        return len(self._pending_syncpoints) == 0
        
    def collect_tasks(self):
        """
        Returns the ticks of all tasks which are ready.
        """
        ticks = set()
        while True:
            tick = self.consume_ready_task()
            if tick is not None:
                ticks.add(tick)
            else:
                break
        return ticks
        
    def consume_ready_task(self):
        """
        Returns the next task which is ready or `None`.
        """
        
        def check_tick_against_sync_point(tick):
            if not self._pending_syncpoints:
                return True # no more sync points
            next_sync_point = self._pending_syncpoints[0]
            
            if tick < next_sync_point:
                # `tick` is not a sync_point and must run
                # before the next sync_point.
                return True
            elif tick == next_sync_point:
                # `tick` is the sync_point
                if self._syncpoint_run_last and self._pending_ticks[0] < next_sync_point:
                    # There are still unfinished tasks that have to run
                    # before it.
                    return False
                else:
                    # It is time to run the sync_point.
                    return True
            else: # tick > next_sync_point
                # has to wait til the sync_point completed
                return False
                
        tick = next(iter(self._queue), None)
        if tick is not None:
            
            if not check_tick_against_sync_point(tick):
                return None
            
            props = self.g.get_task_properties(tick)
            if props.get(self._collected_prop, False):
                raise ValueError("Task %s became ready twice." % tick)
            self.g.set_task_property(tick, self._collected_prop, True)
            self._queue.remove(tick)
        return tick
    
    def was_collected(self, tick):
        """
        Returns if the given tick was collected.
        """
        props = self.g.get_task_properties(tick)
        return props.get(self._collected_prop, False)
        
    
    def add_task(self, tick, task, properties={}):
        properties = dict(properties)
        properties[self._count_prop] = 0
        properties[self._ready_prop] = 0
        properties[self._collected_prop] = False
        self.g.add_task(tick, task, properties=properties)
        self._consider(tick)
        self._pending_ticks.add(tick)
        if properties.get("syncpoint", False):
            self._pending_syncpoints.add(tick)
        
        
    def remove_task(self, tick):
        self.g.remove_task(tick)
        
        if tick in self._queue:
            self._queue.remove(tick)
        if tick in self._pending_syncpoints:
            self._pending_syncpoints.remove(tick)
        if tick in self._pending_ticks:
            self._pending_ticks.remove(tick)
        

    def connect(self, source, dest):
        self.g.connect(source, dest)
        
        if not self._port_filter(self, dest.tick, dest.port):
            return
    
        source_props = self.g.get_task_properties(source.tick)
        dest_props = self.g.get_task_properties(dest.tick)
    
        self.g.set_task_property(dest.tick, self._count_prop, dest_props[self._count_prop] + 1)
        
        if source.port in source_props["out_data"]:
            self.g.set_task_property(dest.tick, self._ready_prop, dest_props[self._ready_prop] + 1)

        self._consider(dest.tick)
            
    def disconnect(self, source, dest):
        self.g.disconnect(source, dest)
            
        if not self._port_filter(self, dest.tick, dest.port):
            return
            
        source_props = self.g.get_task_properties(source.tick)
        dest_props = self.g.get_task_properties(dest.tick)
    
        self.g.set_task_property(dest.tick, self._count_prop, dest_props[self._count_prop] - 1)
        
        if source.port in source_props["out_data"]:
            self.g.set_task_property(dest.tick, self._ready_prop, dest_props[self._ready_prop] - 1)
            
        self._consider(dest.tick)
            
    def set_task_property(self, tick, key, value):
        retval = AbstractGraphDecorator.set_task_property(self, tick, key, value)
        if key == "syncpoint":
            if not value and tick in self._pending_syncpoints:
                self._pending_syncpoints.remove(tick)
            if value:
                self._pending_syncpoints.add(tick)
        self._consider(tick)
        return retval
            
    def set_output_data(self, tick, outputs):
        self.g.set_output_data(tick, outputs)
        
        if tick in self._pending_syncpoints:
            self._pending_syncpoints.remove(tick)
        if tick in self._pending_ticks:
            self._pending_ticks.remove(tick)
        
        for source, dest in self.g.get_out_connections(tick):
            if source.port in outputs:
                
                if not self._port_filter(self, dest.tick, dest.port):
                    continue

                dest_props = self.get_task_properties(dest.tick)
                self.set_task_property(dest.tick, self._ready_prop, dest_props[self._ready_prop] + 1)
                self._consider(dest.tick)
         
    def _consider(self, tick):
        props = self.get_task_properties(tick)
        
        if props.get(self._collected_prop, False):
            # Already collected
            return False
        
        if props[self._count_prop] == props[self._ready_prop]:
            if self._property_filter(self, tick, props):
                should_be_in_queue = True
            else:
                should_be_in_queue = False
        else:
            should_be_in_queue = False
          
        if should_be_in_queue:
            self._queue.add(tick)
        elif tick in self._queue:
            self._queue.remove(tick)
        
class ReadyDecorator(AbstractReadyDecorator):
    """
    Wrapper around a `DataGraphDecorator` that keeps track of the
    tasks which are ready for execution.
    """
    
    def __init__(self, g):
        
        def port_filter(g, tick, port):
            return True
        
        def property_filter(g, tick, props):
            # unrefined tasks are not ready
            task = g.get_task(tick)
            if getattr(task, "refiner_ports", []):
                return props.get("refined", False)
            else:
                return True
        
        AbstractReadyDecorator.__init__(self, g, "ready", port_filter, property_filter)
        
    def collect_ready_tasks(self):
        """
        Returns all tasks for which there is data for all input connections.
        
        Only the tasks that entered this state since the last call are returned.
        
        Changes to the graph that would make a task not ready that was already
        returned by this function are not allowed.
        """
        return self.collect_tasks()
    
    def was_ready_collected(self, tick):
        """
        Returns if the tick was returned by :meth:`collect_ready_tasks`.
        """
        return self.was_collected(tick)
    
            
class RefineDecorator(AbstractReadyDecorator):
    """
    Wrapper around a `DataGraphDecorator` that keeps track of the
    tasks which are ready for refinement.
    """
    
    def __init__(self, g):
        def port_filter(g, tick, port):
            task = g.get_task(tick)
            return port in getattr(task, "refiner_ports", [])
        
        def property_filter(g, tick, props):
            return True
        
        AbstractReadyDecorator.__init__(self, g, "ref", port_filter, property_filter, syncpoint_run_last=False)
        
    def collect_refine_tasks(self):
        """
        Returns all tasks for which there is data for all input connections.
        
        Only the tasks that entered this state since the last call are returned.
        
        Changes to the graph that would make a task not ready that was already
        returned by this function are not allowed.
        """
        ticks = set()
        for tick in self.collect_tasks():
            props = self.get_task_properties(tick)
            if props[self._count_prop] > 0:
                ticks.add(tick)
        return ticks

    def was_refine_collected(self, tick):
        """
        Returns if the tick was returned by :meth:`collect_refine_tasks`.
        """
        return self.will_be_refined(tick) and self.was_collected(tick)
    
    def will_be_refined(self, tick):
        """
        Returns if the tick will be refined at all.
        """
        props = self.get_task_properties(tick)
        return props[self._count_prop] > 0


# Copyright (C) 2015 Stefan C. Mueller

import unittest

from pydron_dataflow.dataflow import tasks
from pydron_dataflow.dataflow.graph import G, T, C, FINAL_TICK, START_TICK, Endpoint, Tick
from pydron_dataflow.interpreter import traverser
from twisted.internet import defer
from twisted.python import failure
import twistit

TICK1 = START_TICK + 1
TICK2 = START_TICK + 2

class TestTraverser(unittest.TestCase):
    
    def setUp(self):
        
        self.refine = []
        self.ready = []
        
        def refine_task_callback(g, tick, task, inputs):
            self.assertEqual(traverser.TaskState.REFINING, self.target.get_task_state(tick))
            d = defer.Deferred()
            self.refine.append((g, tick, task, inputs, d))
            return d
        
        def ready_task_callback(g, tick, task, inputs):
            self.assertEqual(traverser.TaskState.EVALUATING, self.target.get_task_state(tick))
            d = defer.Deferred()
            self.ready.append((g, tick, task, inputs, d))
            return d
        
        self.target = traverser.Traverser(refine_task_callback, ready_task_callback)
        
    def next_refine(self):
        try:
            return self.refine.pop(0)
        except IndexError:
            return None
    
    def next_ready(self):
        try:
            return self.ready.pop(0)
        except IndexError:
            return None
        
    def test_get_task_state_waiting_for_inputs(self):
        g = G(
            T(1, tasks.ConstTask(None)),
            C(1, "value", 2, "in"),
            T(2, "task"),
            C(2, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {})
        self.assertEqual(traverser.TaskState.WAITING_FOR_INPUTS, self.target.get_task_state(TICK2))
        
    def test_get_task_state_evaluating(self):
        g = G(
            T(1, tasks.ConstTask(None)),
            C(1, "value", 2, "in"),
            T(2, "task"),
            C(2, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {})
        self.next_ready()[-1].callback(traverser.EvalResult({"value":"Hello"}))
        self.assertEqual(traverser.TaskState.EVALUATING, self.target.get_task_state(TICK2))
        
    def test_get_task_state_evaluated(self):
        g = G(
            T(1, tasks.ConstTask(None)),
            C(1, "value", 2, "in"),
            T(2, "task"),
            C(2, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {})
        self.next_ready()[-1].callback(traverser.EvalResult({"value":"Hello"}))
        self.assertEqual(traverser.TaskState.EVALUATED, self.target.get_task_state(TICK1))
        
    def test_get_task_state_refining(self):
        task = MockTask("in")
        g = G(
            C(START_TICK, "a", 1, "in"),
            T(1, task),
            C(1, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {"a": "refinedata"})
        
        self.assertEqual(traverser.TaskState.REFINING, self.target.get_task_state(TICK1))

    def test_get_task_state_waiting_for_refiner(self):
        task = MockTask("in")
        g = G(
            T(1, tasks.ConstTask(None)),
            C(1, "value", 2, "in"),
            T(2, task),
            C(2, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {"a": "refinedata"})
        
        self.assertEqual(traverser.TaskState.WAITING_FOR_REFINE_INPUTS, self.target.get_task_state(TICK2))

    def test_no_inputs_task_ready(self):
        g = G(
            T(1, tasks.ConstTask(None)),
            C(1, "value", FINAL_TICK, "retval")
        )
        self.target.execute(g, {})
        self.assertEqual((TICK1, tasks.ConstTask(None), {}), self.next_ready()[1:-1])
        self.assertEqual(None, self.next_ready())
        
    def test_injest_result(self):
        g = G(
            T(1, tasks.ConstTask(None)),
            C(1, "value", 2, "in"),
            T(2, "task"),
            C(2, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {})
        self.next_ready()[-1].callback(traverser.EvalResult({"value":"Hello"}))
        self.assertEqual((TICK2, "task", {"in":"Hello"}), self.next_ready()[1:-1])
        
    def test_finish(self):
        g = G(
            T(1, tasks.ConstTask(None)),
            C(1, "value", FINAL_TICK, "retval")
        )
        d = self.target.execute(g, {})
        self.next_ready()[-1].callback(traverser.EvalResult({"value":"Hello"}))
        
        outputs = extract(d)
        self.assertEqual({"retval":"Hello"}, outputs)
        
        
    def test_finish_nomoretasks(self):
        g = G(
            T(1, tasks.ConstTask(None)),
            C(1, "value", FINAL_TICK, "retval")
        )
        self.target.execute(g, {})
        
        self.next_ready()[-1].callback({"value":"Hello"})
        
        self.assertIsNone(self.next_ready())
        
    def test_eval_fail(self):
        g = G(
            T(1, tasks.ConstTask(None)),
            C(1, "value", FINAL_TICK, "retval")
        )
        d = self.target.execute(g, {})
        self.next_ready()[-1].errback(failure.Failure(MockError()))
        
        
        f = twistit.extract_failure(d)
        self.assertTrue(f.check(traverser.EvaluationError))
        self.assertTrue(f.value.cause.check(MockError))
        self.assertEqual(Tick.parse_tick(1), f.value.tick)
    
    def test_refine_called(self):
        task = MockTask("in")
        g = G(
            C(START_TICK, "a", 1, "in"),
            T(1, task),
            C(1, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {"a": "refinedata"})
        
        self.assertEqual((TICK1, task, {"in": "refinedata"}), self.next_refine()[1:-1])
        
    def test_eval_not_before_refine(self):
        task = MockTask("in")
        g = G(
            C(START_TICK, "a", 1, "in"),
            T(1, task),
            C(1, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {"a": "refinedata"})
        self.assertEqual(None, self.next_ready())
        
    def test_refine_only_once(self):
        task = MockTask("in")
        g = G(
            C(START_TICK, "a", 1, "in"),
            T(1, task),
            C(1, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {"a": "refinedata"})
        
        self.next_refine()[-1].callback(None)
        self.assertEqual(None, self.next_refine())
        
    def test_eval_after_refine(self):
        task = MockTask("in")
        g = G(
            C(START_TICK, "a", 1, "in"),
            T(1, task),
            C(1, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {"a": "refinedata"})
        
        self.next_refine()[-1].callback(None)
        self.assertEqual((TICK1, task, {"in":"refinedata"}), self.next_ready()[1:-1])
        
    def test_refine_fail(self):
        task = MockTask("in")
        g = G(
            C(START_TICK, "a", 1, "in"),
            T(1, task),
            C(1, "out", FINAL_TICK, "retval")
        )
        d = self.target.execute(g, {"a": "refinedata"})
        
        self.next_refine()[-1].errback(failure.Failure(MockError()))
        
        f = twistit.extract_failure(d)
        self.assertTrue(f.check(traverser.RefineError))
        self.assertTrue(f.value.cause.check(MockError))
        self.assertEqual(Tick.parse_tick(1), f.value.tick)
        
    def test_refine_removes_task(self):
        task = MockTask("in")
        g = G(
            C(START_TICK, "a", 1, "in"),
            T(1, task),
            C(1, "out", FINAL_TICK, "retval")
        )
        self.target.execute(g, {"a": "refinedata"})
        
        refine_g, _, _, _, refine_d = self.next_refine()
        refine_g.disconnect(Endpoint(START_TICK, "a"), Endpoint(TICK1, "in"))
        refine_g.disconnect(Endpoint(TICK1, "out"), Endpoint(FINAL_TICK, "retval"))
        refine_g.remove_task(TICK1)
        refine_d.callback(None)
        
        self.assertEqual(None, self.next_refine())
        
        
class MockError(Exception):
    pass
        
class MockTask(object):
    def __init__(self, *refiner_ports):
        self.refiner_ports = set(refiner_ports)

def extract(d):
    """
    Returns the value the given deferred passes to the callback or raises
    the exception passed to errback. If the deferred has no result yet
    a ValueError is raised.
    """
    if not d.called:
        raise ValueError("Deferred not yet called")
    
    failure = []
    value = []
    
    def fail(f):
        failure.append(f)
    
    def success(v):
        value.append(v)
        
    d.addCallbacks(success, fail)
    
    if failure:
        failure[0].raiseException()
    else:
        return value[0]


# Copyright (C) 2015 Stefan C. Mueller

from pydron_dataflow.dataflow import graph
from pydron_dataflow.interpreter import graphdecorator

from twisted.internet import defer
from twisted.internet.defer import CancelledError
from twisted.python import failure

import logging
import enum
logger = logging.getLogger(__name__)

class TaskState(enum.Enum):
    """
    State of a task during traversal.
    
    Tasks that don't need refinement won't ever be in
    states `WAITING_FOR_REFINE_INPUTS` and `REFINING`.
    
    Not all tasks will go through all states as they
    may be deleted during a refinement or have all inputs
    ready right from the beginning. Refinement (if needed)
    is guaranteed to happen before evaluation.
    """
    
    #: Task needs refinement, but we don't have
    #: the refiner inputs yet. The task has not
    #: been evaluated yet.
    WAITING_FOR_REFINE_INPUTS = 1,
    
    #: Task is currently being refined.
    #: The task has not been evaluated yet.
    REFINING = 2,
    
    #: Task needs to be evaluated, but we don't
    #: have the the inputs yet. The task
    #: is already refined or does not need
    #: refinement.
    WAITING_FOR_INPUTS = 3,
    
    #: Task is currently being evaluated.
    #: The task is refined or does not need
    #: refinement.
    EVALUATING = 4,
    
    #: Task has been evaluated.
    #: The task is refined or does not need
    #: refinement.
    EVALUATED = 5

class EvalResult(object):
    """
    Result of a task evaulation.
    """
    def __init__(self, result, duration=None, datasizes=None, transfer_results=None):
        """
        :param result: Either a :class:`failure.Failure` or a `dict` with out-port name
         to value map, where value is typically a :class:`worker.ValueId`.
         
        :param duration: Evalation time in seconds.
        
        :param datasizes: `dict` with out-port to byte-count mapping. If a port is missing
            then that implies that it's value cannot be pickled.
        
        :param transfer_results: `dict` with port to :class:`worker.TransmissionResult` mapping for
            inputs that had to be transferred first.
        """
        self.result = result
        self.duration = duration
        self.datasizes = datasizes
        self.transfer_results = transfer_results
        
    def __repr__(self):
        return "EvalResult(%r, %r, %r, %r)" % (self.result, self.duration, self.datasizes, self.transfer_results)
    

class EvaluationError(Exception):
    """
    Error while evaluating a task.
    
    :attr:`graph` Refined graph.
    :attr:`tick` Tick of the task that failed to evaluate.
    :attr:`cause` A :class`twisted.Failure` instance with the cause of the evaulation error.
    """
    def __init__(self, graph, tick, cause):
        Exception.__init__(self, "Evaluation of tick %r failed: %s" %(tick, cause.getTraceback()))
        self.graph = graph
        self.tick = tick
        self.cause = cause
        
class RefineError(Exception):
    """
    Error while refining a task.
    
    :attr:`graph` Refined graph.
    :attr:`tick` Tick of the task that failed to refine.
    :attr:`cause` A :class`twisted.Failure` instance with the cause of the refine error.
    """
    def __init__(self, graph, tick, cause):
        Exception.__init__(self, "Refinement of %r failed: %s" %(tick, cause.getTraceback()))
        self.graph = graph
        self.tick = tick
        self.cause = cause

class Traverser(object):
    """
    Traverses the graph, performs graph refinement and invokes a callback
    to ask the scheduler to execute tasks.
        
    The actual record-keeping is all done in :mod:`graphdecorator`. This is
    mostly glue code to give a nicer API for the scheduler and 
    `ScheduledCallable` based on deferreds.
    """
    
    def __init__(self, refine_task_callback, ready_task_callback):
        """
        :param ready_task_callback: function which is invoked when a task
            becomes ready for execution.
            
            It is invoked with four arguments:
             * `graph` The graph we are traversing with all the decorators around it.
             * `tick` Tick of the task.
             * `task` Task object, equal to `graph.get_task(tick)`
             * `inputs` `dict` that maps input ports to values (or rather value references).
            
            The callable has to return a deferred that maps output ports to values (or rather value references). 
            If it failbacks the traversal aborts with that failure. 
            
            The returned deferred might get cancelled if the result isn't required.
            
        :param refine_task_callback: Function which is invoked when a task
            becomes ready for refinement. The signature is the same as for`ready_task_callback`,
            the return value is ignored.
            
            The returned deferred might get cancelled if the result isn't required.
        """

                
        self._refine_task_callback = refine_task_callback
        self._ready_task_callback = ready_task_callback
        self._result = defer.Deferred(self._cancel)
        
        #: maps tick to the deferred that we passed to `ready_task_callback`.
        #: we use this to cancel the operation if needed.
        #: Contains `None` during the invocation of the callback (before we get
        #: the deferred back).
        self._pending_ready_deferreds = {}
        
        #: maps tick to the deferred that we passed to `refine_task_callback`.
        #: we use this to cancel the operation if needed.
        #: Contains `None` during the invocation of the callback (before we get
        #: the deferred back).
        self._pending_refine_deferreds = {}
        
        #: Set if a failure occured that should stop the execution and
        #: be passed on to the user.
        self._caught_failure = None
        
        self._graph = None
        self._started = False
        self._finished = False
        
        
    def get_graph(self):
        """
        Returns the graph which is refined as traversal is progressing.
        
        Once :meth:`execute` has finished, this function can be used
        to get the fully refined graph.
        """
        return self._graph
    
    def get_task_state(self, tick):
        """
        Returns the :class:`TaskState` describing the current state of the given
        task.
        """
        ready_collected = self._graph.was_ready_collected(tick)
        refine_collected = self._graph.was_refine_collected(tick)
        will_be_refined = self._graph.will_be_refined(tick)

        if tick in self._pending_refine_deferreds:
            return TaskState.REFINING
        if tick in self._pending_ready_deferreds:
            return TaskState.EVALUATING
        
        if will_be_refined:
            if refine_collected:
                if ready_collected:
                    return TaskState.EVALUATED
                else:
                    return TaskState.WAITING_FOR_INPUTS
            else:
                return TaskState.WAITING_FOR_REFINE_INPUTS

        else:
            if ready_collected:
                return TaskState.EVALUATED
            else:
                return TaskState.WAITING_FOR_INPUTS

    def execute(self, g, inputs):
        """
        Runs the given graph with the given inputs.
        
        :param g: Graph to traverse. Won't be changed.
        
        :param inputs: `dict` with port to `ValueRef` map.
        
        :returns: deferred `dict` with port to `ValueRef` map for the outputs.
            The deferred can be cancelled.
        """
        
        if self._started:
            raise ValueError("Traverser can only be started once.")
        self._started = True
        
        self._graph = graph.Graph()
        self._graph = graphdecorator.DataGraphDecorator(self._graph)
        self._graph = graphdecorator.RefineDecorator(self._graph)
        self._graph = graphdecorator.ReadyDecorator(self._graph)
        
        # copy the graph into g
        for tick in g.get_all_ticks():
            self._graph.add_task(tick, g.get_task(tick), g.get_task_properties(tick))
        for tick in g.get_all_ticks() + [graph.FINAL_TICK]:
            for source, dest in g.get_in_connections(tick):
                self._graph.connect(source, dest)

        # ingest graph inputs
        self._graph.set_output_data(graph.START_TICK, inputs)
        
        self._iterate()
        
        return self._result

    def _iterate(self):
        
        if self._finished:
            return
        
        # Stop traversing on error
        if self._caught_failure:
            self._finished = True
            self._result.errback(self._caught_failure)
            return

        # Lets see what work is ready to do
        ready_for_execution = self._graph.collect_ready_tasks()
        ready_for_refine = self._graph.collect_refine_tasks()
        
        # If the final tick is ready for execution, we have the graph
        # outputs and can finish traversing.
        if graph.FINAL_TICK in ready_for_execution:
            self._finished = True
            graph_outputs = {dest.port : self._graph.get_data(source) for source, dest in self._graph.get_in_connections(graph.FINAL_TICK)}
            self._result.callback(graph_outputs)
            return
        
        # Pass the refine jobs to the scheduler.
        for tick in ready_for_refine:
            task = self._graph.get_task(tick)
            inputs = {dest.port : self._graph.get_data(source) for source, dest in self._graph.get_in_connections(tick) if dest.port in task.refiner_ports}
            
            # We add a property to the task so that we can later check if the refine operation
            # removed the task and replaced it with a different one.
            token = object()
            self._graph.set_task_property(tick, "refine_token", token)
            
            self._pending_refine_deferreds[tick] = None
            d = self._refine_task_callback(self._graph, tick, task, inputs)
            self._pending_refine_deferreds[tick] = d
            
            def on_success(value, tick, token):
                del self._pending_refine_deferreds[tick] 
                
                # Mark the task as refined. Using the token to see if the task is still
                # in the graph.
                try:
                    token_found = self._graph.get_task_properties(tick)["refine_token"] == token
                except KeyError:
                    token_found = False
                if token_found:
                    self._graph.set_task_property(tick, "refined", True)
                
                self._iterate()
        
            def on_fail(fail, tick):
                del self._pending_refine_deferreds[tick]
                try:
                    raise RefineError(self.get_graph(), tick, fail)
                except:
                    fail = failure.Failure()
                return self._handle_failure(fail)
            
            d.addCallbacks(on_success, on_fail, callbackArgs=(tick,token), errbackArgs=(tick,))
            
            def unhandled(failure):
                logger.error(failure.getTraceback())
                
            d.addErrback(unhandled)
        
        # Pass the evaluation jobs to the scheduler.
        for tick in ready_for_execution:
            task = self._graph.get_task(tick)
            inputs = {dest.port : self._graph.get_data(source) for source, dest in self._graph.get_in_connections(tick)}
            
            self._pending_ready_deferreds[tick] = None
            d = self._ready_task_callback(self._graph, tick, task, inputs)
            self._pending_ready_deferreds[tick] = d
        
            def on_success(evalresult, tick):
                del self._pending_ready_deferreds[tick]
                
                if evalresult.duration is not None:
                    self._graph.set_task_property(tick, "eval_time", evalresult.duration)
                
                if isinstance(evalresult.result, dict):
                    outputs = evalresult.result
                    
                    if evalresult.datasizes is not None:
                        self._graph.set_task_property(tick, "datasizes", evalresult.datasizes)
                    
                    self._graph.set_output_data(tick, outputs)
                    
                        
                elif isinstance(evalresult.result, failure.Failure):
                    logger.debug("Evaluation of %r has caused exception: " % tick + evalresult.result.getTraceback())
                    
                    try:
                        raise EvaluationError(self.get_graph(), tick, evalresult.result)
                    except:
                        fail = failure.Failure()
                    
                    return self._handle_failure(fail)
                
                else:
                    raise ValueError("Unexpected result for tick %r:%r" % (tick, evalresult.result))
                
                self._iterate()
            
            def on_fail(fail, tick):
                del self._pending_ready_deferreds[tick]
            
                try:
                    raise EvaluationError(self.get_graph(), tick, fail)
                except:
                    fail = failure.Failure()
            
                return self._handle_failure(fail)
            
            d.addCallbacks(on_success, on_fail, callbackArgs=(tick,), errbackArgs=(tick,))

            def unhandled(failure):
                logger.error(failure.getTraceback())
                
            d.addErrback(unhandled)


    def _cancel(self, d):
        """
        Cancel the traversal with a CancelledError.
        """
        if self._finished:
            return
        try:
            raise CancelledError()
        except:
            self._caught_failure = failure.Failure()
            self._iterate()

    
    def _handle_failure(self, failure):
        if self._finished:
            if failure.check(CancelledError):
                return None # we cancel pending tasks on finish.
            else:
                # Cannot really do anything with this failure 
                # since we've already called back to the user.
                return failure 
        else:
            # Lets remember the failure and iterate. This
            # will cause the traversal to finish and
            # all other pending operations to be cancelled.
            self._caught_failure = failure
            self._iterate()
            return None


# Copyright (C) 2015 Stefan C. Mueller
import logging

logger = logging.getLogger(__name__)


class AbstractTask(object):
    
    # Input ports required for refining.
    # We won't refine if the attribute is missing.
    #refiner_ports = set()
    
    # Maps input port to a function that takes one
    # argument. If the attribute exists then
    # this function is invoked with the value
    # and the return value is passed to `refine`
    # instead of the actual value.
    # The function must be defined in a module
    # so that it can be pickled.
    # The idea of this is to reduce the amount
    # of data that has to be transferred back
    # to the scheduler for refinement.
    #refiner_reducer = {}
    
    def input_ports(self):
        raise NotImplementedError("abstract")
    
    def output_ports(self):
        raise NotImplementedError("abstract")
    
    def subgraphs(self):
        return tuple()
    
    def evaluate(self, inputs):
        raise NotImplementedError("abstract")
    
    def __eq__(self, other):
        return self.__dict__ == other.__dict__
    def __ne__(self, other):
        return not self == other
    def __hash__(self):
        return 1


class ConstTask(AbstractTask):
    """
    Evaluates to a constant value.
    """

    def __init__(self, value):
        self.value = value

    def input_ports(self):
        return set()

    def output_ports(self):
        return {"value"}

    def evaluate(self, inputs):
        return {'value': self.value}

    def __repr__(self):
        return "ConstTask(" + repr(self.value) + ")"

    def __eq__(self, other):
        return self.value == other.value


# Copyright (C) 2014 Stefan C. Mueller

import os.path
from setuptools import setup, find_packages

if os.path.exists('README.rst'):
    with open('README.rst') as f:
        long_description = f.read()
else:
    long_description = None
    
setup(
    name = 'pydron_dataflow',
    version = '0.2.6',
    description='Dataflow framework to define and traverse directional graphs.',
    long_description=long_description,
    author='Ivo Nussbaumer',
    author_email='ivo.nussbaumer@fhnw.ch',
    url='https://github.com/i4Ds/pydron-dataflow',
    packages = find_packages(),
    install_requires = ['astor>=0.4',
                        'frozendict>=2.0',
                        'sortedcontainers>=0.9.5']
)



# Copyright (C) 2015 Stefan C. Mueller

import threading
import sys
from frozendict import frozendict

class Tick(object):
    """
    Tick objects represent a 'time' during sequential execution of the
    corresponding python code.
    
    Tick instances can be compared and are immutable.
    """
    
    def __init__(self, elements, loopmask):
        """
        Do not use. Instead derive from other tick, such as `START_TICK`.
        """
        if len(elements) < 2:
            raise ValueError()
        if elements[0] > 1:
            raise ValueError("cannot have tick after FINAL_TICK")
        if elements[0] == 1 and elements != (1,0):
            raise ValueError("cannot have tick after FINAL_TICK")
        if not isinstance(elements, tuple) or not isinstance(loopmask, tuple):
            raise TypeError()
        if len(loopmask) != len(elements):
            raise ValueError("Loopmask does not match elements")
        for e in elements:
            if (sys.version_info > (3, 0)):
                if not isinstance(e, int):
                    raise TypeError()
                if e < 0:
                    raise ValueError("elements must be non-negative.")
            else:
                if not isinstance(e, (int, long)):
                    raise TypeError()
                if e < 0:
                    raise ValueError("elements must be non-negative.")
        self._elements = elements
        self._loopmask = loopmask
    
    def mark_loop_iteration(self):
        """
        Marks this tick as the base tick of a loop iteration.
        """
        return Tick(self._elements, self._loopmask[:-1] + (True,))
    
    @property
    def nonloop_elements(self):
        """
        Returns a tuple of integers representing those elements of the tick that are not loop iterations.
        If two ticks share the same `nonloop_elements` then they represent the same tick (task) but 
        potentially from two different loop iterations.
        """
        return tuple(self._elements[i] for i in range(len(self._elements)) if not self._loopmask[i])
    
    @property
    def loop_elements(self):
        """
        Returns a tuple of integers representing those elements of the tick that are loop iterations.
        The first element is the outer most loop. The last element the inner most loop.
        """
        return tuple(self._elements[i] for i in range(len(self._elements)) if self._loopmask[i])
        
    def __add__(self, count):
        """
        Returns the clock `count` cycles after this one. 
        Does not change the loop mask.
        """
        return Tick(self._elements[:-1] + (self._elements[-1] + count,), self._loopmask)
    
    def __lshift__(self, other):
        """
        Returns a clock that is after `other` and before `other + 1` while
        retaining the order of all clocks that have been left-shifted by
        `other`.
        """
        if other == FINAL_TICK:
            raise ValueError("Cannot move beyond FINAL_TICK")
        return Tick(other._elements + self._elements[1:], other._loopmask + self._loopmask[1:])
    
    def __rshift__(self, positions):
        """
        
        """
        if not isinstance(positions, int):
            raise ValueError("Right shifting a tick expects an integer.")
        return Tick(self._elements[:-positions], self._loopmask[:-positions])
    
    def __cmp__(self, other):
        return cmp(self._elements, other._elements)

    def __lt__(self, other):
        return self._elements < other._elements

    def __eq__(self, other):
        return self._elements == other._elements

    def __hash__(self):
        return hash(self._elements)
    
    def __repr__(self):
        if self == START_TICK:
            return "START_TICK"
        elif self == FINAL_TICK:
            return "FINAL_TICK"
        else:
            length = len(self._elements) 
            if length == 2:
                return self._repr_element(1)
            else:
                if (sys.version_info > (3, 0)):
                    return "(" + ", ".join(map(self._repr_element, list(range(1, length)))) + ")"
                return "(" + ", ".join(map(self._repr_element, range(1, length))) + ")"

    def _repr_element(self, index):
        if self._loopmask[index]:
            return "*%s" % self._elements[index]
        else:
            return str(self._elements[index])
        
    @staticmethod   
    def _parse_element(elt):
        elt = elt.strip()
        loop = elt.startswith("*")
        if loop:
            elt = elt[1:]
        return int(elt), loop
            
    @staticmethod   
    def parse_tick(tick):
        if isinstance(tick, Tick):
            return tick
        elif isinstance(tick, str):
            if tick == "START_TICK":
                return START_TICK
            elif tick == "FINAL_TICK":
                return FINAL_TICK
            else:
                elements = [0]
                loopmask = [False]
                elts = tick.split(",")
                for elt in elts:
                    element, loop  = Tick._parse_element(elt)
                    elements.append(element)
                    loopmask.append(loop)
                
                return Tick(tuple(elements), tuple(loopmask))
            
        elif isinstance(tick, tuple):
            return Tick((0,) + tick, (False,) * (len(tick) + 1))
        else:
            return Tick((0, tick), (False, False))

START_TICK = Tick((0,0), (False, False))
FINAL_TICK = Tick((1,0), (False, False))

class GraphObserver(object):
    """
    Base class for graph observers.
    Ignores all events by default.
    """
    
    def connected(self, source, dest):
        pass
    
    def disconnected(self, source, dest):
        pass
    
    def task_added(self, tick, task, properties):
        pass
    
    def task_removed(self, tick, task):
        pass
    
    def task_property_changed(self, tick, key, value):
        pass

class Endpoint(object):
    """
    Port of a specific node.
    """
    def __init__(self, tick, port):
        assert isinstance(tick, Tick)
        assert isinstance(port, str)
        assert port is not None
        self.tick = tick
        self.port = port

    def __repr__(self):
        return "Endpoint(%s, %s)" % (repr(self.tick), repr(self.port))

    def __eq__(self, other):
        return self.tick == other.tick and self.port == other.port

    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __hash__(self):
        return hash(self.tick) + 7*hash(self.port)


class _TaskNode(object):
    """
    Used internally by :class:`Graph`.
    """
    
    def __init__(self, task, properties):
        self.task = task
        self.properties = frozendict(properties)
        self.in_connections = {}
        self.out_connections = set()
        
    def set_property(self, key, value):
        try:
            self.properties = self.properties.copy(**{key: value})
        except Exception:
            self.properties = frozendict(self.properties | {key: value})
        
    def __eq__(self, other):
        if self.task != other.task:
            return False
        
        if (sys.version_info > (3, 0)):
            myproperties = {k:v for k,v in self.properties.items() if not k.startswith("_")}
            otherproperties = {k:v for k,v in other.properties.items() if not k.startswith("_")}
        else:
            myproperties = {k:v for k,v in self.properties.iteritems() if not k.startswith("_")}
            otherproperties = {k:v for k,v in other.properties.iteritems() if not k.startswith("_")}
        
        if myproperties != otherproperties:
            return False
        
        if self.in_connections != other.in_connections:
            return False
        
        if self.out_connections != other.out_connections:
            return False
        
        return True
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        raise NotImplementedError()   
        
class _Connection(object):
    """
    Used internally by :class:`Graph`.
    """
    
    def __init__(self, source, dest):
        assert isinstance(source, Endpoint)
        assert isinstance(dest, Endpoint)
        self.source = source
        self.dest = dest

    def __eq__(self, other):
        return (self.source == other.source and
                self.dest == other.dest)
        
    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __hash__(self):
        return (1 *hash(self.source) + 
                3 * hash(self.dest))
        
    def __repr__(self):
        return "(%s, %s)" % (repr(self.source), repr(self.dest))


class Graph(object):
    """
    A graph consists of tasks and data-flow connections between them.
    Each task is identified by a :class:`Tick` which represents the time
    at which that task would execute in a sequential program.
    
    Each task has input and output ports. Each port has an identifier
    object which must be unique for the task. This class makes no
    assumtion on the identifier object other than comparability.
    
    An output port may be connected to zero or more input ports.
    Each input port is connected to one output port, but may be disconnected.
    
    Causality of connections must be kept: the output tick must be older 
    than the input tick. This also forbids cicles.
    
    `START_TICK` and `FINAL_TICK` are already in the graph and serve as dummy-tasks.
    One can make connections to them to mark data passed into the graph and
    out of it.
    
    The graph is observable. Observers are objects with the following methods:
    
    .. py:function:: added_task(graph, tick, task, properties)
    
    .. py:function:: removed_task(graph, tick)
    
    .. py:function:: connected(graph, source, dest)
    
    .. py:function:: disconnected(graph, source, dest)
    
    The methods are called *after* the graph has been changed. During those calls
    the graph may *not* be changed with the exception of task properties. Exceptions
    thrown in those methods leave the graph in an undefined state. Changes to
    task properties are not reported.
    """
    


    def __init__(self):
        self._ticks = {START_TICK:_TaskNode(None, {}),
                       FINAL_TICK:_TaskNode(None, {})}
        self._observers = []

    def add_task(self, tick, task, properties={}):
        """
        Adds a task at the specified tick.
        
        The new task will be unconnected.
        
        :param tick: Time at which this task executes.
        :param properties: Properties of the task.
        """
        if tick == START_TICK:
            raise ValueError("START_TICK is reserved for inputs of the graph.")
        if tick == FINAL_TICK:
            raise ValueError("FINAL_TICK is reserved for outputs of the graph.")
        
        tasknode = _TaskNode(task, properties)
        if tick in self._ticks:
            raise ValueError("Tick already has a task")
        self._ticks[tick] = tasknode
        
        self._fire_task_added(tick, task, properties)

    def remove_task(self, tick):
        """
        Removes the task at the given tick from the graph.
        The task must be unconnected.
        """
        if not tick in self._ticks:
            raise ValueError("No task which that tick")
        if self.get_in_connections(tick) or self.get_out_connections(tick):
            raise ValueError("Task is connected")
        del self._ticks[tick]
        
        self._fire_task_removed(tick)
            
    def connect(self, source, dest):
        """
        Create a connection between two tasks.
        
        :param source: Source endpoint
        :param dest: Dest endpoint
        :param throw_exists: Throw exception if connection already exists.
        
        :returns: `True` if the connection was created, False if the
        connection already exists and `throw_exists` is `False`.
        """
        
        if dest.tick < source.tick:
            raise ValueError("Destination executes before source")
        
        if source.tick == FINAL_TICK:
            raise ValueError("FINAL_TICK can only have input ports")
        
        if dest.tick == START_TICK:
            raise ValueError("START_TICK can only have output ports")
        
        
        conn = _Connection(source, dest)
        
        source_task = self._ticks[source.tick]
        dest_task = self._ticks[dest.tick]
        
        if dest.port in dest_task.in_connections:
            if dest_task.in_connections[dest.port] == conn:
                # connection already exists
                raise ValueError("The connection %s already exists" % repr(conn))
            else:
                raise ValueError("The destination port is already connected: %s " % repr(conn))
        
        if conn in source_task.out_connections:
            raise ValueError("This connection already exists")
        
        dest_task.in_connections[dest.port] = conn
        source_task.out_connections.add(conn)

        self._fire_connected(source, dest)
        return True
    
    def disconnect(self, source, dest):
        """
        Remove a connection between two tasks.
        
        :param source: Source endpoint
        :param dest: Dest endpoint
        """
        source_task = self._ticks[source.tick]
        dest_task = self._ticks[dest.tick]
        
        conn = _Connection(source, dest)
        
        if conn not in source_task.out_connections:
            raise ValueError("This connection does not exists")
        
        del dest_task.in_connections[dest.port]
        source_task.out_connections.remove(conn)
        
        self._fire_disconnected(source, dest)
    
    def get_all_ticks(self):
        """
        Returns the ticks of all tasks.
        """
        def gen():
            if (sys.version_info > (3, 0)):
                for tick in self._ticks.keys():
                    if tick != START_TICK and tick != FINAL_TICK:
                        yield tick
            else:
                for tick in self._ticks.iterkeys():
                    if tick != START_TICK and tick != FINAL_TICK:
                        yield tick
        return list(gen())
    
    def get_task(self, tick):
        """
        Returns the task at the given tick.
        """
        return self._ticks[tick].task
    
    def get_task_properties(self, tick):
        """
        Returns the properties for the task executed at `tick`.
        """
        return self._ticks[tick].properties
        
    def set_task_property(self, tick, key, value):
        """
        Change a property of the task at the given tick.
        """
        task = self._ticks[tick]
        task.set_property(key, value)
        
        self._fire_task_property_changed(tick, key, value)
        
    def get_in_connections(self, tick):
        """
        Returns `(source, dest)` tuples for each input
        connection of the task at the given tick.
        """
        def gen():
            task = self._ticks[tick]
            if (sys.version_info > (3, 0)):
                for _, conn in task.in_connections.items():
                    yield (conn.source, conn.dest)
            else:
                for _, conn in task.in_connections.iteritems():
                    yield (conn.source, conn.dest)
        return list(gen())
                
    def get_out_connections(self, tick):
        """
        Returns `(source_endpoint, dest_endpoint)` tuples for each output
        connection of the task at the given tick.
        """
        def gen():
            task = self._ticks[tick]
            for conn in task.out_connections:
                yield (conn.source, conn.dest)
        return list(gen())
    
    def subscribe(self, observer):
        """
        Registers an observer.
        """
        self._observers.append(observer)
        
    def unsubscribe(self, observer):
        """
        Unregisters an observer.
        """
        self._observers.remove(observer)
    
    def _fire_task_added(self, tick, task, properties):
        for obs in self._observers:
            obs.task_added(tick, task, properties)
            
    def _fire_task_removed(self, tick):
        for obs in self._observers:
            obs.task_removed(tick)
            
    def _fire_connected(self, source, dest):
        for obs in self._observers:
            obs.connected(source, dest)
            
    def _fire_disconnected(self, source, dest):
        for obs in self._observers:
            obs.disconnected(source, dest)
            
    def _fire_task_property_changed(self, tick, key, value):
        for obs in self._observers:
            obs.task_property_changed(tick, key, value)  
              
    def __repr__(self):
        stack = getattr(Graph, "_repr_stack", [])
        Graph._repr_stack = stack
        if any(s is self for s in stack):
            return "<recursive graph>"
        stack.append(self)
        
        def indent(s):
            lines = s.splitlines()
            return "\n   ".join(lines)
            
        def gen_arguments():
            if (sys.version_info > (3, 0)):
                for tick in sorted(self._ticks.keys()):
                    if tick == START_TICK:
                        continue
                    tasknode = self._ticks[tick]
                    for conn in tasknode.in_connections.values():
                        yield "%s.%s -> %s.%s" % (conn.source.tick, conn.source.port, conn.dest.tick, conn.dest.port)
                    
                    if tick == FINAL_TICK:
                        continue
                    if tasknode.properties:
                        yield repr(tick) + ": " + indent(repr(tasknode.task)) + " " + indent(repr(dict(tasknode.properties))) 
                    else:
                        yield repr(tick) + ": " + indent(repr(tasknode.task))
            else:
                for tick in sorted(self._ticks.iterkeys()):
                    if tick == START_TICK:
                        continue
                    tasknode = self._ticks[tick]
                    for conn in tasknode.in_connections.itervalues():
                        yield "%s.%s -> %s.%s" % (conn.source.tick, conn.source.port, conn.dest.tick, conn.dest.port)
                    
                    if tick == FINAL_TICK:
                        continue
                    if tasknode.properties:
                        yield repr(tick) + ": " + indent(repr(tasknode.task)) + " " + indent(repr(dict(tasknode.properties))) 
                    else:
                        yield repr(tick) + ": " + indent(repr(tasknode.task))
        
        if len(self.get_all_ticks()) == 0 and len(self.get_in_connections(FINAL_TICK)) == 0:
            s = "{}"
        else:
            s = "{\n   " + "\n   ".join(gen_arguments()) + "\n}"
        stack.pop()
        return s

    def __eq__(self, other):
        with _comparison_context(self, other) as already_comparing:
            if already_comparing:
                return True
            
            for tick in sorted(self._ticks):
                
                mytask = self._ticks[tick]
                if tick not in other._ticks:
                    return False
                othertask = other._ticks[tick]
                
                if mytask != othertask:
                    return False
                
            return True
    
    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __hash__(self):
        raise NotImplementedError("Not supported")




class _comparison_context(object):
    """
    Graphs may be recursive. We use a bit of thread-local magic
    to break the loop.
    """
    th = threading.local()
    
    def __init__(self, graph_a, graph_b):
        self.graph_a = graph_a
        self.graph_b = graph_b
        
    def getid(self):
        return (id(self.graph_a), id(self.graph_b))

    def __enter__(self):
        if not hasattr(self.th, "stack"):
            self.th.stack = []
            
        already_comparing = self.getid() in self.th.stack
        self.th.stack.append(self.getid())
        return already_comparing

    def __exit__(self, *exec_info):
        self.th.stack.pop()
    
    @classmethod
    def comparing(cls, graph_a, graph_b):
        return cls.th.already_compared.get((id(graph_a), id(graph_b)), None)


    




class graph_factory(object):
    factory_thread_local = threading.local()
    
    def __enter__(self):
        self.factory_thread_local.named_graphs = {}
    def __exit__(self, *exec_info):
        del self.factory_thread_local.named_graphs
        return False


def G(*commands):

    if commands and isinstance(commands[0], str):
        name = commands[0]
        commands = commands[1:]
    else:
        name = None
        
    if name:
        g = graph_factory.factory_thread_local.named_graphs.get(name, Graph())
        graph_factory.factory_thread_local.named_graphs[name] = g
    else:
        g = Graph()
    
    tasks = [cmd for cmd in commands if isinstance(cmd, T)]
    connections = [cmd for cmd in commands if isinstance(cmd, C)]
    
    for task in tasks:
        g.add_task(task.tick, task.task, task.properties)
        
    for conn in connections:
        g.connect(conn.source, conn.dest)
    
    return g
            
class T (object):
    def __init__(self, tick, task, properties={}):
        self.tick = Tick.parse_tick(tick)
        self.task = task
        self.properties = properties
        
class C(object):
    def __init__(self, source_tick, source_port, dest_tick, dest_port):        
        self.source = Endpoint(Tick.parse_tick(source_tick), source_port)
        self.dest = Endpoint(Tick.parse_tick(dest_tick), dest_port)
        
